// O valor dos campos de data não considerem o fuso horário local
process.env.ORA_SDTZ = 'UTC';

const oracledb = require('oracledb');
const path = require('path');
const fs = require('fs');

// eslint-disable-next-line import/no-dynamic-require
const oracleConfig = require(path.resolve(__dirname, '../../../config/OracleConfig'));

const moment = require('moment');
require('moment/locale/pt-br');

const Now = function () {
  return moment().format('Y-MM-DD HH:mm:ss');
};

class OracleService {
  constructor() {
    this.oracledb = oracledb;
    this.connection = {};
  }

  async connect() {
    let customDbname = !!process.argv[2] ? process.argv[2] : "";

    if (await this.checkConnection()) {
      return this.connection;
    }

    if (customDbname.indexOf("ODHEN") != -1){
      customDbname = '';
    }

    let connect = {};

    if(!customDbname) {
      connect = await oracledb.getConnection({
        user: oracleConfig.user,
        password: oracleConfig.password,
        connectString: oracleConfig.connectionString,
      });

      this.connection = connect;
      return connect;
    }

    connect = await oracledb.getConnection({
 	    externalAuth: true,
      connectString: customDbname
    });

    this.connection = connect;
    return connect;
  }

  async fetch(sqlString, params) {
    let sqlPrepare = sqlString;

    if (params) {
      Object.keys(params).forEach((key) => {
        const replace = new RegExp(`:${key}`, 'g');
        if (key.search('DT') == 0 && params[key].length > 9) {
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        } else {
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        }
      });
    }

    if (! await this.checkConnection()) {
      this.connection = await this.connect();

      await this.connection.execute("ALTER SESSION SET TIME_ZONE='UTC'");
    }

    return this.connection.execute(sqlPrepare);
  }

  async fetchAssoc(sqlString, params) {
    let sqlPrepare = sqlString;

    if (params) {
      Object.keys(params).forEach((key) => {
        const replace = new RegExp(`:${key}`, 'g');
        if (key.search('DT') == 0 && params[key].length > 9) {
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        } else {
          sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'`);
        }
      });
    }

    if (! await this.checkConnection()) {
      this.connection = await this.connect();

      await this.connection.execute("ALTER SESSION SET TIME_ZONE='UTC'");
    }
    return this.connection.execute(
      sqlPrepare,
      [],
      {
        outFormat: oracledb.OUT_FORMAT_OBJECT
      }
    );
  }

  async close() {
    await this.connection.close(function (err) {
      if (err) {
        console.error(err.message);
      }
    });
  }

  async commit() {
    if(await this.checkConnection()) {
      await this.connection.commit();
    }
  }

  async rollback() {
    if (await this.checkConnection()) {
      await this.connection.rollback();
    }
  }

  async checkConnection() {
    if (Object.keys(this.connection).length === 0) {
      return false;
    }

    if (typeof this.connection.ping !== 'function') {
      return false;
    }

    return this.connection.ping().then((err) => {
      if(err) {
        this.connection.close();
        return false;
      }
      return true;
    });
  }

  async execProcedure(sqlString, params) {
    let sqlPrepare = sqlString;

    if (! await this.checkConnection()) {
      this.connection = await this.connect();
    }
    return await this.connection.execute(sqlPrepare, params);
  }

  async execStoredProcedure(name, params) {
    if (name) {
      let paramNames = Object.keys(params);

      if (paramNames.length) {
        paramNames = paramNames.map((param) => `:${param}`).join(', ');
      } else {
        paramNames = '';
      }

      const sql = `BEGIN
        ${name}(${paramNames});
      END;`;

      if (! await this.checkConnection()) {
        this.connection = await this.connect();
      }

      return this.connection.execute(sql, params);
    }

    return null;
  }

  async getFieldType(queryResult, fieldName) {}

  async execute(sqlString, params, outParams = [], extendedMetaData = false, connection = null) {
    let sqlPrepare = sqlString;
    let paramsKeys = params ? Object.keys(params) : null;
    let anonymousBlock = false;
    let colPrepare = {};

    paramsKeys = params
      ? paramsKeys.sort(function (a1, a2) {
          return ~(a1.length < a2.length);
        })
      : [];

    if (paramsKeys) {
      paramsKeys.forEach((key) => {
        try {
          const replace = new RegExp(`:${key}(\\s|,|\\)|$)`, 'g');
          const comma = sqlPrepare.match(replace)[0].includes(',') ? ',' : '';
          const closeParenthesis = sqlPrepare.match(replace)[0].includes(')') ? ')' : '';

          if (params[key] == null) {
            sqlPrepare = sqlPrepare.replace(replace, `null${comma}${closeParenthesis}`);
          } else if (typeof params[key] === 'object') {
            anonymousBlock = true;
            const colName = key;

            const regExp = new RegExp('"', 'g');
            let paramData = String.fromCharCode(...params[key].data);
            paramData = paramData.replace(regExp, "'");
            colPrepare[colName] = paramData;
          } else if (typeof params[key] === 'string') {
            const regExp = new RegExp("'", 'g');
            params[key] = params[key].replace(regExp, "''");
            sqlPrepare = sqlPrepare.replace(replace, `'${params[key]}'${comma}${closeParenthesis}`);
          } else {
            sqlPrepare = sqlPrepare.replace(replace, `${params[key]}${comma}${closeParenthesis}`);
          }
        } catch (error) {
          // erros que podem acontecer é a chave aparecer no update não como condição e sim para substituir no update
        }
      });
    }
    if (! await this.checkConnection()) {
      this.connection = await this.connect();
    }

    sqlPrepare = sqlPrepare.replace(/TIMESTAMP null/gi, 'null');
    sqlPrepare = sqlPrepare.replace(/TIMESTAMP 'Invalid date'/gi, 'null');

    const dbConnection = connection || this.connection;

    if (anonymousBlock) {
      const connect = dbConnection
        .execute(`BEGIN ${sqlPrepare}; END;`, colPrepare)
        .then((result) => {
          const newResult = result;
          newResult.connection = this.connection;

          return newResult;
        })
        .catch((error) => {
          this.logSql(sqlString, sqlPrepare);
          throw error;
        });
      return connect;
    }
    try {
      oracledb.fetchAsString = [ oracledb.CLOB ];
      const dbConnectionResult = await dbConnection.execute(sqlPrepare, outParams, {
        extendedMetaData,
      });

      dbConnectionResult.connection = dbConnection;
      return dbConnectionResult;
    } catch (error) {
      const debug = true;

      if (debug) {
        this.logSql(sqlString, sqlPrepare);
      }

      this.saveSalesImportException(error, sqlPrepare, params);

      throw error;
    }
  }

  getNewConnection() {
    let customDbname = !!process.argv[2] ? process.argv[2] : "";

    if (customDbname.indexOf("ODHEN") != -1){
      customDbname = '';
    }

    if(!customDbname) {
      return this.oracledb.getConnection({
        user: oracleConfig.user,
        password: oracleConfig.password,
        connectString: oracleConfig.connectionString,
      });
    }

    return this.oracledb.getConnection({
 	    externalAuth: true,
      connectString: customDbname
    });
  }

  logSql(sqlString, sqlPrepare) {
    let fileToPutData = '';
    if (fs.existsSync('./sqlData.sql')) {
      fileToPutData = fs.readFileSync('./sqlData.sql');
    }

    fileToPutData += sqlPrepare + '\r\n';

    let fileToPutString = '';
    if (fs.existsSync('./sqlString.sql')) {
      fileToPutString = fs.readFileSync('./sqlString.sql');
    }

    fileToPutString += sqlString + '\r\n';

    fs.writeFileSync('./sqlData.sql', fileToPutData);
    fs.writeFileSync('./sqlString.sql', fileToPutString);
  }

  async saveSalesImportException(exception, sqlStatement, sqlParams) {
    try {
      let tableName = '';

      const statment = sqlStatement.trim().toUpperCase();
      let testValue = statment.match(/BEGIN/) || statment.match(/DECLARE/);
      if (!testValue) {
        testValue = statment.match(/INTO/) || statment.match(/SET/);
      }

      if (testValue) {
        if (testValue[0] === 'INTO') {
          const inicial = statment.indexOf('INTO') + 5;
          const final = statment.indexOf('(');
          tableName = statment.substring(inicial, final).trim();
        } else {
          const inicial = statment.indexOf('SET');
          tableName = statment.substring(inicial, 6).trim();
        }
      }

      if (tableName !== 'ERROIMPORTAVENDA' && tableName) {
        const primaryKey = `
          SELECT COLUMN_NAME, POSITION FROM ALL_CONS_COLUMNS WHERE CONSTRAINT_NAME = 'PK_${tableName}'  GROUP BY COLUMN_NAME, POSITION ORDER BY POSITION
        `;

        const primaryKeyResult = await this.execute(primaryKey);
        if (tableName === 'VENDA') {
          primaryKeyResult.rows[2][0] = 'NRSEQVENDAMSDE';
        }
        let key = '';

        if (primaryKeyResult.rows) {
          for (const columnName of primaryKeyResult.rows) {
            if (sqlParams[columnName[0]] || sqlParams[`P_${columnName[0]}`]) {
              key =
                `${key + columnName[0]}=${sqlParams[columnName[0]]},` ||
                `${sqlParams[`P_${columnName[0]}`]},`;
            } else {
              key = '';
              break;
            }
          }
        }

        const params = [];
        const properties = Object.keys(sqlParams);

        for (const i of properties) {
          const value = `${i} = ${sqlParams[i]}`;
          params.push(value);
        }

        const dsParams = params.join();
        if (!key) {
          key = dsParams.substring(0, 4000);
        } else {
          key = key.substring(0, key.length - 1).trim();
        }

        const statmentSql = sqlStatement.substring(0, 4000);
        const insertStatement = `
          INSERT INTO ERROIMPORTAVENDA
          (CHAVE, TABELA, DSINSTRUCAOSQL, DSPARAMS, DSERRO, DTERRO)
          VALUES
          (:P_CHAVE, :P_TABELA, SUBSTR(:P_DSINSTRUCAOSQL, 1, 2000), SUBSTR(:P_DSPARAMS, 1, 2000),:P_DSERRO, TIMESTAMP :P_DTERRO)
        `;

        const errorParams = {
          P_CHAVE: key,
          P_TABELA: tableName,
          P_DSINSTRUCAOSQL: statmentSql.trim(),
          P_DSPARAMS: dsParams.substring(0, 4000),
          P_DSERRO: exception['message'],
          P_DTERRO: Now(),
        };

        await this.execute(insertStatement, errorParams);
      }
    } catch (error) {
      console.log(error);
    }
  }
}

module.exports = new OracleService();